"""
Helper for ASAM ODS HTTP API conI session

Example::

    from odsbox.con_i import ConI

    with ConI(
        url="http://localhost:8087/api",
        auth=("sa", "sa")
    ) as con_i:
        units = con_i.query_data({"AoUnit": {}})

"""

from __future__ import annotations

import logging
import os
from typing import List, Tuple

import requests
import requests.auth
from google.protobuf.json_format import MessageToJson
from google.protobuf.message import Message
from pandas import DataFrame

import odsbox.proto.ods_pb2 as ods
from odsbox.datamatrices_to_pandas import to_pandas
from odsbox.jaquel import jaquel_to_ods
from odsbox.model_cache import ModelCache
from odsbox.transaction import Transaction
from odsbox.security import Security


class ConI:
    """
    This is a helper to hold an ASAM ODS HTTP API ConI session.

    Example::

        from odsbox.con_i import ConI

        with ConI(
            url="http://localhost:8087/api",
            auth=("sa", "sa")
        ) as con_i:
            units = con_i.query_data({"AoUnit": {}})

    """

    __log: logging.Logger = logging.getLogger(__name__)
    __default_http_headers: dict[str, str] = {
        "Content-Type": "application/x-asamods+protobuf",
        "Accept": "application/x-asamods+protobuf",
    }

    def __init__(
        self,
        url: str = "http://localhost:8080/api",
        auth: requests.auth.AuthBase | Tuple[str, str] = ("sa", "sa"),
        context_variables: ods.ContextVariables | dict | None = None,
        verify_certificate: bool = True,
        load_model: bool = True,
        allow_redirects: bool = False,
    ):
        """
        Create a session object keeping track of ASAM ODS session URL named `conI`.

        Example::

            from odsbox.con_i import ConI

            # basic auth
            with ConI(
                url="http://localhost:8087/api",
                auth=("sa", "sa")
            ) as con_i:
                units = con_i.query_data({"AoUnit": {}})

        Example::

            import requests
            from odsbox.con_i import ConI

            class BearerAuth(requests.auth.AuthBase):
                def __init__(self, token):
                    self.token = token
                def __call__(self, r):
                    r.headers["authorization"] = "Bearer " + self.token
                    return r

            # bearer auth
            with ConI(
                url="http://localhost:8087/api",
                auth=BearerAuth("YOUR_BEARER_TOKEN")
            ) as con_i:
                units = con_i.query_data({"AoUnit": {}})


        :param str url: base url of the ASAM ODS API of a given server. An example is "http://localhost:8080/api".
        :param requests.auth.AuthBase auth: An auth object to be used for the used requests package.
            For basic auth `("USER", "PASSWORD")` can be used.
        :param ods.ContextVariables | dict | None context_variables: If context variables are necessary for the
            connection they are passed here. It defaults to None.
        :param bool verify_certificate: If no certificate is provided for https insecure access can be enabled.
            It defaults to True.
        :param bool load_model: If the model should be read after connection is established. It defaults to True.
        :param bool allow_redirects: If redirects should be allowed in requests calls. It defaults to False.
        :raises requests.HTTPError: If connection to ASAM ODS server fails.
        """
        self.__session: requests.Session | None = None
        self.__con_i: str | None = None
        self.__security: Security | None = None
        self.__mc: ModelCache | None = None
        self.__allow_redirects: bool = allow_redirects

        session = requests.Session()
        session.auth = auth
        session.verify = verify_certificate

        _context_variables = None
        if isinstance(context_variables, ods.ContextVariables):
            _context_variables = context_variables
        else:
            _context_variables = ods.ContextVariables()
            if isinstance(context_variables, dict):
                for key, value in context_variables.items():
                    _context_variables.variables[key].string_array.values.append(value)

        response = session.post(
            url + "/ods",
            data=_context_variables.SerializeToString(),
            timeout=60.0,
            headers=self.__default_http_headers,
            allow_redirects=self.__allow_redirects,
        )
        if 201 == response.status_code:
            con_i = response.headers["location"]
            self.__log.debug("ConI: %s", con_i)
            self.__session = session
            self.__con_i = con_i
        self.check_requests_response(response)
        if load_model:
            # lets cache the model
            self.model_read()

    def __del__(self):
        if self.__session is not None:
            self.logout()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.logout()

    def con_i_url(self) -> str:
        """
        Get the ASAM ODS session URL used to work with this session.

        :return str: The ASAM ODS session URL
        """
        if self.__con_i is None:
            raise ValueError("ConI already closed")
        return self.__con_i

    def logout(self):
        """
        Close the attached session at the ODS server by calling delete on the session URL
        and closing the requests session.

        :raises requests.HTTPError: If delete the ASAM ODS session fails.
        """
        if self.__session is not None:
            if self.__con_i is None:
                raise ValueError("ConI already closed")
            response = self.__session.delete(
                self.__con_i,
                timeout=60.0,
                headers={"Accept": "application/x-asamods+protobuf"},
                allow_redirects=self.__allow_redirects,
            )
            self.__session.close()
            self.__session = None
            self.__con_i = None
            self.__security = None
            self.__mc = None
            self.check_requests_response(response)

    def query_data(
        self,
        query: str | dict | ods.SelectStatement,
        enum_as_string: bool = False,
        date_as_timestamp: bool = False,
        **kwargs,
    ) -> DataFrame:
        """
        Query ods server for content and return the results as Pandas DataFrame

        :param str | dict | ods.SelectStatement query: Query given as JAQueL query (dict or str)
            or as an ASAM ODS SelectStatement.
        :param bool enum_as_string: columns of type DT_ENUM or DS_ENUM are returned as int values.
                                    If this is set to True the model_cache is used to map the int values
                                    to the corresponding string values.
        :param bool date_as_timestamp: columns of type DT_DATE or DS_DATE are returned as string.
                                       If this is set to True the strings are converted to pandas Timestamp.
        :param kwargs: additional arguments passed to `to_pandas`.
        :raises requests.HTTPError: If query fails.
        :return DataFrame: The DataMatrices as Pandas.DataFrame. The columns are named as `ENTITY_NAME.ATTRIBUTE_NAME`.
            `IsNull` values are not marked invalid.
        """
        data_matrices = (
            self.data_read(query) if isinstance(query, ods.SelectStatement) else self.data_read_jaquel(query)
        )
        return to_pandas(
            data_matrices,
            model_cache=self.mc,
            enum_as_string=enum_as_string,
            date_as_timestamp=date_as_timestamp,
            **kwargs,
        )

    def model(self) -> ods.Model:
        """
        Get the cache ODS server model. This model will return the cached
        application model related to your session.

        :return ods.Model: The application model of the ASAM ODS server.
        """
        return self.mc.model()

    def data_read_jaquel(self, jaquel: str | dict) -> ods.DataMatrices:
        """
        Query ods server for content.

        :param str | dict  jaquel: Query given as JAQueL query (dict or str).
        :raises requests.HTTPError: If query fails.
        :return ods.DataMatrices: The DataMatrices representing the result.
            It will contain one ods.DataMatrix for each returned entity type.
        """
        _, ods_query = jaquel_to_ods(self.model(), jaquel)
        return self.data_read(ods_query)

    def data_read(self, select_statement: ods.SelectStatement) -> ods.DataMatrices:
        """
        Query ods server for content.

        :param ods.SelectStatement  select_statement: Query given as ASAM ODS SelectStatement.
        :raises requests.HTTPError: If query fails.
        :return ods.DataMatrices: The DataMatrices representing the result.
            It will contain one ods.DataMatrix for each returned entity type.
        """
        if not isinstance(select_statement, ods.SelectStatement):
            raise TypeError(f"data_read expects 'ods.SelectStatement', got '{type(select_statement).__name__}'")
        response = self.ods_post_request("data-read", select_statement)
        return_value = ods.DataMatrices()
        return_value.ParseFromString(response.content)
        return return_value

    def data_create(self, data: ods.DataMatrices) -> List[int]:
        """
        Create new ASAM ODS instances or write bulk data.

        :param ods.DataMatrices data: Matrices containing columns for instances to be created.
        :raises requests.HTTPError: If creation fails.
        :return List[int]: list of ids created from your request.
        """
        if not isinstance(data, ods.DataMatrices):
            raise TypeError(f"data_create expects 'ods.DataMatrices', got '{type(data).__name__}'")
        response = self.ods_post_request("data-create", data)
        return_value = ods.DataMatrices()
        return_value.ParseFromString(response.content)
        return list(return_value.matrices[0].columns[0].longlong_array.values)

    def data_update(self, data: ods.DataMatrices) -> None:
        """
        Update existing instances.

        :param ods.DataMatrices data: Matrices containing columns for instances to be updated.
            The `id` column is used to identify the instances to be updated.
        :raises requests.HTTPError: If update fails.
        """
        if not isinstance(data, ods.DataMatrices):
            raise TypeError(f"data_update expects 'ods.DataMatrices', got '{type(data).__name__}'")
        self.ods_post_request("data-update", data)

    def data_delete(self, data: ods.DataMatrices) -> None:
        """
        Delete existing instances.

        :param ods.DataMatrices data: Matrices containing columns for instances to be deleted.
            The `id` column is used to identify the instances to be deleted.
        :raises requests.HTTPError: If delete fails.
        """
        if not isinstance(data, ods.DataMatrices):
            raise TypeError(f"data_delete expects 'ods.DataMatrices', got '{type(data).__name__}'")
        self.ods_post_request("data-delete", data)

    def data_copy(self, copy_request: ods.CopyRequest) -> ods.Instance:
        """
        Copy an Instance and its related children.

        :param ods.CopyRequest copy_request: Define instance to be copied.
        :raises requests.HTTPError: If copy fails.
        :return ods.Instance: Newly created instance
        """
        if not isinstance(copy_request, ods.CopyRequest):
            raise TypeError(f"data_copy expects 'ods.CopyRequest', got '{type(copy_request).__name__}'")
        response = self.ods_post_request("data-copy", copy_request)
        return_value = ods.Instance()
        return_value.ParseFromString(response.content)
        return return_value

    def n_m_relation_read(self, identifier: ods.NtoMRelationIdentifier) -> ods.NtoMRelatedInstances:
        """
        Read n-m relations for a defined instance.

        :param ods.NtoMRelationIdentifier identifier: identify n to m relation to be read.
        :raises requests.HTTPError: If read fails.
        :return ods.NtoMRelatedInstances: Return n to m related instances that were queried.
        """
        if not isinstance(identifier, ods.NtoMRelationIdentifier):
            raise TypeError(
                f"n_m_relation_read expects 'ods.NtoMRelationIdentifier', got '{type(identifier).__name__}'"
            )
        response = self.ods_post_request("n-m-relation-read", identifier)
        return_value = ods.NtoMRelatedInstances()
        return_value.ParseFromString(response.content)
        return return_value

    def n_m_relation_write(self, related_instances: ods.NtoMWriteRelatedInstances) -> None:
        """
        Update, delete or create n-m relations for given instance pairs.

        :raises requests.HTTPError: If write fails.
        :param ods.NtoMWriteRelatedInstances related_instances: related instances to be
            updated, deleted or created.
        """
        if not isinstance(related_instances, ods.NtoMWriteRelatedInstances):
            raise TypeError(
                f"n_m_relation_write expects 'ods.NtoMWriteRelatedInstances', got '{type(related_instances).__name__}'"
            )
        self.ods_post_request("n-m-relation-write", related_instances)

    def transaction(self) -> Transaction:
        """
        Open a transaction object to be used in a with clause

        Example::

            with con_i.transaction() as transaction:
                # do writing
                transaction.commit()

        :raises requests.HTTPError: If creation of transaction fails.
        :return Transaction: transaction object that will abort automatically if commit is not called.
        """
        return Transaction(self)

    def transaction_create(self) -> None:
        """
        Open a transaction for writing.
        :raises requests.HTTPError: If creation of transaction fails.
        """
        self.ods_post_request("transaction-create")

    def transaction_commit(self) -> None:
        """
        Commit transaction created before.
        :raises requests.HTTPError: If creation of transaction fails.
        """
        self.ods_post_request("transaction-commit")

    def transaction_abort(self) -> None:
        """
        Abort transaction created before.
        :raises requests.HTTPError: If creation of transaction fails.
        """
        self.ods_post_request("transaction-abort")

    def valuematrix_read(self, request: ods.ValueMatrixRequestStruct) -> ods.DataMatrices:
        """
        Read bulk data from a submatrix or measurement.
        Submatrix access can also be done using data-read.

        :param ods.ValueMatrixRequestStruct request: Define measurement or submatrix to
            create ASAM ODS ValueMatrix for.
        :raises requests.HTTPError: If ValueMatrix access fails.
        :return ods.DataMatrices: DataMatrices containing the bulk data for the request.
        """
        if not isinstance(request, ods.ValueMatrixRequestStruct):
            raise TypeError(f"valuematrix_read expects 'ods.ValueMatrixRequestStruct', got '{type(request).__name__}'")
        response = self.ods_post_request("valuematrix-read", request)
        return_value = ods.DataMatrices()
        return_value.ParseFromString(response.content)
        return return_value

    def model_read(self) -> ods.Model:
        """
        Read the model from server and update cached version.

        :raises requests.HTTPError: If model read fails.
        :return ods.Model: The application model of the server.
        """
        response = self.ods_post_request("model-read")
        model = ods.Model()
        model.ParseFromString(response.content)
        self.__mc = ModelCache(model)
        return model

    def model_update(self, model_parts: ods.Model, update_model: bool = True) -> None:
        """
        Update application model content. This method is used to modify existing items or
        create new ones.

        :param ods.Model model_parts: parts of the model to be updated or created.
        :param bool update_model: determine if the model cache of this ConI instance should be
            updated by reading the whole model again. It defaults to True.
        :raises requests.HTTPError: If model update fails.
        """
        if not isinstance(model_parts, ods.Model):
            raise TypeError(f"model_update expects 'ods.Model', got '{type(model_parts).__name__}'")
        self.ods_post_request("model-update", model_parts)
        if update_model:
            # cache again if successfully changed
            self.model_read()

    def model_delete(self, model_parts: ods.Model, update_model: bool = True) -> None:
        """
        Delete application model content.

        :param ods.Model model_parts: define model parts to be deleted.
        :param bool update_model: determine if the model cache of this ConI instance should be
            updated by reading the whole model again. It defaults to True.
        :raises requests.HTTPError: If model update fails.
        """
        if not isinstance(model_parts, ods.Model):
            raise TypeError(f"model_delete expects 'ods.Model', got '{type(model_parts).__name__}'")
        self.ods_post_request("model-delete", model_parts)
        if update_model:
            # cache again if successfully changed
            self.model_read()

    def model_check(self) -> None:
        """
        Check if stored application model is consistent.

        :raises requests.HTTPError: If model contains errors.
        """
        self.ods_post_request("model-check")

    def basemodel_read(self) -> ods.BaseModel:
        """
        Read the ODS base model version used by the server.

        :raises requests.HTTPError: If reading base model fails.
        :return ods.BaseModel: used server base model.
        """
        response = self.ods_post_request("basemodel-read")
        base_model = ods.BaseModel()
        base_model.ParseFromString(response.content)
        return base_model

    def asampath_create(self, instance: ods.Instance) -> ods.AsamPath:
        """
        Create an persistent string representing the instance.

        :param ods.Instance instance: Instance to be get AsamPath for.
        :raises requests.HTTPError: If creation fails.
        :return ods.AsamPath: The AsamPath that represents the instance.
        """
        if not isinstance(instance, ods.Instance):
            raise TypeError(f"asampath_create expects 'ods.Instance', got '{type(instance).__name__}'")
        response = self.ods_post_request("asampath-create", instance)
        return_value = ods.AsamPath()
        return_value.ParseFromString(response.content)
        return return_value

    def asampath_resolve(self, asam_path: ods.AsamPath) -> ods.Instance:
        """
        Use the persistent string to get back the instance.

        :param ods.AsamPath asam_path: AsamPath to be resolved.
        :raises requests.HTTPError: If path could not be resolved.
        :return ods.Instance: Instance represented by AsamPath.
        """
        if not isinstance(asam_path, ods.AsamPath):
            raise TypeError(f"asampath_resolve expects 'ods.AsamPath', got '{type(asam_path).__name__}'")
        response = self.ods_post_request("asampath-resolve", asam_path)
        return_value = ods.Instance()
        return_value.ParseFromString(response.content)
        return return_value

    def context_read(self, pattern_or_filter: ods.ContextVariablesFilter | str = "*") -> ods.ContextVariables:
        """
        Read the session context variables.

        :param ods.ContextVariablesFilter | str pattern_or_filter: Context variable filter as str
            or ContextVariablesFilter. It defaults to "*" to return all variables.
        :raises requests.HTTPError: If something went wrong.
        :return ods.ContextVariables: ContextVariables where the name matches the filter.
        """
        context_variables_filter = (
            pattern_or_filter
            if isinstance(pattern_or_filter, ods.ContextVariablesFilter)
            else ods.ContextVariablesFilter(pattern=pattern_or_filter)
        )
        response = self.ods_post_request("context-read", context_variables_filter)
        return_value = ods.ContextVariables()
        return_value.ParseFromString(response.content)
        return return_value

    def context_update(self, context_variables: ods.ContextVariables):
        """
        Set context variables for current session. This will set context variables for the given session.
        If new session is created they will fall back to their default.

        :param ods.ContextVariables context_variables: ContextVariables to be set or updated.
        :raises requests.HTTPError: If something went wrong.
        """
        if not isinstance(context_variables, ods.ContextVariables):
            raise TypeError(f"context_update expects 'ods.ContextVariables', got '{type(context_variables).__name__}'")
        self.ods_post_request("context-update", context_variables)

    def password_update(self, password_update: ods.PasswordUpdate) -> None:
        """
        Update the password of the defined user.

        :param ods.PasswordUpdate password_update: Defines for which user the password should eb updated.
        :raises requests.HTTPError: If something went wrong.
        """
        if not isinstance(password_update, ods.PasswordUpdate):
            raise TypeError(f"password_update expects 'ods.PasswordUpdate', got '{type(password_update).__name__}'")
        self.ods_post_request("password-update", password_update)

    def file_access(self, file_identifier: ods.FileIdentifier) -> str:
        """
        Get file access URL for file content.

        :param ods.FileIdentifier file_identifier: Define content to be accessed.
                                                   Might be an AoFile or a DT_BLOB attribute.
        :raises requests.HTTPError: If something went wrong.
        :raises ValueError: If no file location provided by server.
        :return str: The server file URL.
        """
        if not isinstance(file_identifier, ods.FileIdentifier):
            raise TypeError(f"file_access expects 'ods.FileIdentifier', got '{type(file_identifier).__name__}'")
        response = self.ods_post_request("file-access", file_identifier)
        server_file_url = response.headers.get("location")
        if server_file_url is None:
            raise ValueError("No file location provided by server!")
        return server_file_url

    def file_access_download(
        self,
        file_identifier: ods.FileIdentifier,
        target_file_or_folder: str,
        overwrite_existing: bool = False,
        default_filename: str = "download.bin",
    ) -> str:
        """
        Read file content from server.

        :param ods.FileIdentifier file_identifier: Define content to be read. Might be an AoFile or a DT_BLOB attribute.
        :param str target_file_or_folder: Path to save the file content to. If pointing to an existing folder. Original
                                          filename will be used. Full path is returned.
        :param bool overwrite_existing: If existing files should be overwritten. It defaults to False.
        :param str default_filename: Default filename if no filename is provided by server.
                                     It defaults to "download.bin".
        :raises requests.HTTPError: If something went wrong.
        :raises FileExistsError: If file already exists and 'overwrite_existing' is False.
        :raises ValueError: If no open session.
        :return str: file path of saved file.
        """
        if not isinstance(file_identifier, ods.FileIdentifier):
            raise TypeError(
                f"file_access_download expects 'ods.FileIdentifier', got '{type(file_identifier).__name__}'"
            )
        server_file_url = self.file_access(file_identifier)

        if self.__session is None:
            raise ValueError("No open session!")
        file_response = self.__session.get(
            server_file_url,
            headers={
                "Accept": "application/octet-stream, application/x-asamods+protobuf, */*",
            },
            allow_redirects=self.__allow_redirects,
        )
        self.check_requests_response(file_response)

        target_file_path = target_file_or_folder
        if os.path.isdir(target_file_path):
            content_disposition = file_response.headers.get(
                "Content-Disposition", f'attachment; filename="{default_filename}"'
            )
            filename = (
                content_disposition.split("filename=")[1].strip('"')
                if "filename=" in content_disposition
                else default_filename
            )
            target_file_path = os.path.join(target_file_path, filename)

        if not overwrite_existing and os.path.exists(target_file_path):
            raise FileExistsError(f"File '{target_file_path}' already exists and 'overwrite_existing' is False.")

        with open(target_file_path, "wb") as file:
            file.write(file_response.content)

        return target_file_path

    def file_access_upload(
        self,
        file_identifier: ods.FileIdentifier,
        source_file_path: str,
    ) -> None:
        """
        Upload file content to server.

        :param ods.FileIdentifier file_identifier: Define content to be written.
                                                   Might be an AoFile or a DT_BLOB attribute.
        :param str source_file_path: Path to the file to be uploaded.
        :raises requests.HTTPError: If something went wrong.
        :raises FileNotFoundError: If source file was not found.
        :raises ValueError: If no open session.
        """
        if not isinstance(file_identifier, ods.FileIdentifier):
            raise TypeError(f"file_access_upload expects 'ods.FileIdentifier', got '{type(file_identifier).__name__}'")
        if not os.path.isfile(source_file_path):
            raise FileNotFoundError(f"File '{source_file_path}' not found.")

        server_file_url = self.file_access(file_identifier)

        with open(source_file_path, "rb") as file:
            if self.__session is None:
                raise ValueError("No open session!")
            put_response = self.__session.put(
                server_file_url,
                data=file,
                headers={"Content-Type": "application/octet-stream", "Accept": "application/x-asamods+protobuf"},
                allow_redirects=self.__allow_redirects,
            )
            self.check_requests_response(put_response)

    def file_access_delete(
        self,
        file_identifier: ods.FileIdentifier,
    ) -> None:
        """
        Delete file content from server.

        :param ods.FileIdentifier file_identifier: Define content to be deleted.
                                                   Might be an AoFile or a DT_BLOB attribute.
        :raises requests.HTTPError: If something went wrong.
        :raises ValueError: If no open session.
        """
        if not isinstance(file_identifier, ods.FileIdentifier):
            raise TypeError(f"file_access_delete expects 'ods.FileIdentifier', got '{type(file_identifier).__name__}'")
        server_file_url = self.file_access(file_identifier)

        if self.__session is None:
            raise ValueError("No open session!")
        delete_response = self.__session.delete(
            server_file_url,
            headers={"Accept": "application/x-asamods+protobuf"},
            allow_redirects=self.__allow_redirects,
        )
        self.check_requests_response(delete_response)

    def ods_post_request(
        self,
        relative_url_part: str,
        message: Message | None = None,
        timeout: float = 600.0,
        headers: dict[str, str] | None = None,
    ) -> requests.Response:
        """
        Do ODS post call with the given relative URL.

        :param str relative_url_part: url part that is joined to conI URL using `/`.
        :param Message | None message: protobuf message to be send, defaults to None.
        :param float timeout: maximal time to wait for response.
        :raises requests.HTTPError: If status code is not 200 or 201.
        :return requests.Response: requests response if successful.
        """

        if self.__session is None or self.__con_i is None:
            raise ValueError("No open session!")

        response = self.__session.post(
            self.__con_i + "/" + relative_url_part,
            data=message.SerializeToString() if message is not None else None,
            timeout=timeout,
            headers=(headers if headers is not None else self.__default_http_headers),
            allow_redirects=self.__allow_redirects,
        )
        self.check_requests_response(response)
        return response

    @staticmethod
    def check_requests_response(response: requests.Response):
        if response.status_code not in (200, 201):
            response.headers
            if (
                "Content-Type" in response.headers
                and "application/x-asamods+protobuf" == response.headers["Content-Type"]
            ):
                error_info = ods.ErrorInfo()
                error_info.ParseFromString(response.content)
                raise requests.HTTPError(
                    MessageToJson(error_info),
                    response=response,
                )
            response.raise_for_status()

    @property
    def mc(self) -> ModelCache:
        """
        Get the model cache for the current session.

        :return ods.ModelCache: ModelCache object containing the cached application model.
        """
        if self.__mc is None:
            raise ValueError("Model not read! Call model_read() first.")
        return self.__mc

    @property
    def security(self) -> Security:
        """
        Get the security information for the current session.

        :raises requests.HTTPError: If security info retrieval fails.
        :return ods.Security: Security object containing permissions and roles.
        """
        if self.__session is None:
            raise ValueError("No open session!")

        if self.__security is None:
            self.__security = Security(self)
        return self.__security



"""Integration test for ASAM ODS session"""

from _pytest.fixtures import FixtureRequest
from odsbox.con_i import ConI
from odsbox.submatrix_to_pandas import submatrix_to_pandas
from odsbox.jaquel import jaquel_to_ods
from odsbox.security import Security
import odsbox.proto.ods_pb2 as ods
import odsbox.proto.ods_security_pb2 as ods_security

from google.protobuf.json_format import MessageToJson

import logging
import pytest
import os
import tempfile


def __create_con_i(load_model: bool = True) -> ConI:
    """Create a connection session for an ASAM ODS server"""
    return ConI("https://docker.peak-solution.de:10032/api", ("Demo", "mdm"), load_model=load_model)


def test_con_i():
    with __create_con_i() as con_i:
        model = con_i.model_read()
        assert len(model.entities) > 0

        entity = con_i.mc.entity_by_base_name("AoUnit")
        assert entity.base_name.lower() == "aounit"

        con_i.query_data({"AoEnvironment": {}, "$options": {"$rowlimit": 1}})
        con_i.query_data({"AoUnit": {}, "$options": {"$rowlimit": 1}})
        con_i.query_data({"AoMeasurement": {}, "$options": {"$rowlimit": 1}})
        con_i.query_data(
            {"AoMeasurement": {}, "$options": {"$rowlimit": 50}}, date_as_timestamp=True, enum_as_string=True
        )

        r = con_i.query_data(
            {"AoUnit": {}, "$attributes": {"name": 1}, "$options": {"$rowlimit": 1}}, name_separator="::"
        )
        assert f"{entity.name}::" in r.columns[0]


def test_query_data():
    with __create_con_i() as con_i:
        model = con_i.model_read()
        assert len(model.entities) > 0

        assert con_i.query_data({"AoEnvironment": {}, "$options": {"$rowlimit": 1}}).empty is False
        assert con_i.query_data('{"AoEnvironment": {}, "$options": {"$rowlimit": 1}}').empty is False

        entity, select_statement = jaquel_to_ods(con_i.model(), {"AoEnvironment": {}, "$options": {"$rowlimit": 1}})
        assert entity.base_name == "AoEnvironment"
        assert con_i.query_data(select_statement).empty is False

        entity, select_statement = jaquel_to_ods(con_i.model(), '{"AoEnvironment": {}, "$options": {"$rowlimit": 1}}')
        assert entity.base_name == "AoEnvironment"
        assert con_i.query_data(select_statement).empty is False


def test_transaction():
    with __create_con_i() as con_i:
        with con_i.transaction() as transaction:
            transaction.abort()

        with con_i.transaction() as transaction:
            transaction.commit()

        with con_i.transaction() as transaction:
            pass  # automatically abort


@pytest.mark.integration
def test_submatrix_load():
    with __create_con_i() as con_i:
        sm_s = con_i.query_data(
            {
                "AoSubmatrix": {},
                "$options": {"$rowlimit": 1},
                "$attributes": {"id": 1, "measurement.name": 1},
            }
        )

        logging.getLogger().info(sm_s.shape)
        assert sm_s.shape[0] <= 1
        if 1 == sm_s.shape[0]:
            submatrix_id = int(sm_s.iloc[0, 0])  # type: ignore
            assert 0 != submatrix_id
            logging.getLogger().info(submatrix_id)
            submatrix_dataframe = submatrix_to_pandas(con_i, submatrix_id)
            assert submatrix_dataframe is not None


def test_bug_93_outer_join_on_n_relation():
    logging.getLogger().info(
        "In case of outer join the direction is important and must point from n to 1 when generated from jaquel."
    )
    with __create_con_i() as con_i:
        df = con_i.query_data(
            {
                "AoMeasurement": {},
                "$attributes": {
                    "name": 1,
                    "measurement_quantities:OUTER": {"name": 1, "unit:OUTER.name": 1},
                    "submatrices:OUTER.name": 1,
                },
                "$options": {"$rowlimit": 10},
            }
        )
        assert df is not None
        assert df.empty is False
        assert 4 == len(df.columns)


@pytest.mark.integration
def test_values_data_read(request: FixtureRequest):
    test_name = request.node.name
    with __create_con_i() as con_i:
        sm_df = con_i.query_data(
            {
                "AoSubmatrix": {"name": "Profile_02"},
                "$attributes": {"id": 1, "number_of_rows": 1},
                "$options": {"$rowlimit": 1},
            }
        )
        assert sm_df.empty is False
        sm_id: int = sm_df.iat[0, 0]
        sm_number_of_rows: int = sm_df.iat[0, 1]
        logging.getLogger().info("Submatrix Id: %s, number_of_rows:%s", sm_id, sm_number_of_rows)
        assert sm_id is not None
        assert sm_number_of_rows is not None

        lc_info_df = con_i.query_data(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "name": 1, "sequence_representation": 1, "independent": 1},
            }
        )
        assert lc_info_df.empty is False

        lc_info_dms = con_i.data_read_jaquel(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "name": 1, "sequence_representation": 1, "independent": 1},
            }
        )
        assert 1 == len(lc_info_dms.matrices)
        with open(
            os.path.join(tempfile.gettempdir(), f"{test_name}_lc_info.proto.json"), "w", encoding="utf-8"
        ) as json_file:
            json_file.write(MessageToJson(lc_info_dms))

        lc_values_df = con_i.query_data(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "values": 1, "generation_parameters": 1},
                "$options": {"$seqlimit": 10, "$seqskip": 0},
            }
        )
        assert lc_values_df.empty is False

        lc_values_dms = con_i.data_read_jaquel(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "values": 1, "generation_parameters": 1},
                "$options": {"$seqlimit": sm_number_of_rows, "$seqskip": 0},
            }
        )
        assert 1 == len(lc_values_dms.matrices)
        with open(
            os.path.join(tempfile.gettempdir(), f"{test_name}_lc_values.proto.json"), "w", encoding="utf-8"
        ) as json_file:
            json_file.write(MessageToJson(lc_values_dms))


# @pytest.mark.integration
def test_values_valuematrix_read_calculated(request: FixtureRequest):
    test_name = request.node.name
    with __create_con_i() as con_i:
        sm_df = con_i.query_data(
            {
                "AoSubmatrix": {"name": "Profile_02"},
                "$attributes": {"id": 1, "number_of_rows": 1},
                "$options": {"$rowlimit": 1},
            }
        )
        assert sm_df.empty is False
        sm_id: int = sm_df.iat[0, 0]
        sm_number_of_rows: int = sm_df.iat[0, 1]
        logging.getLogger().info("Submatrix Id: %s, number_of_rows:%s", sm_id, sm_number_of_rows)
        assert sm_id is not None
        assert sm_number_of_rows is not None

        lc_entity = con_i.mc.entity_by_base_name("AoLocalColumn")
        assert lc_entity is not None

        lc_info_df = con_i.query_data(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "name": 1, "sequence_representation": 1, "independent": 1},
            }
        )
        assert lc_info_df.empty is False

        lc_info_dms = con_i.data_read_jaquel(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "name": 1, "sequence_representation": 1, "independent": 1},
            }
        )
        assert 1 == len(lc_info_dms.matrices)
        with open(
            os.path.join(tempfile.gettempdir(), f"{test_name}_lc_info.proto.json"), "w", encoding="utf-8"
        ) as json_file:
            json_file.write(MessageToJson(lc_info_dms))

        lc_attributes = [
            con_i.mc.attribute_by_base_name(lc_entity, "name").name,
            con_i.mc.attribute_by_base_name(lc_entity, "values").name,
        ]

        lc_values_dms = con_i.valuematrix_read(
            ods.ValueMatrixRequestStruct(
                aid=con_i.mc.entity_by_base_name("AoSubmatrix").aid,
                iid=sm_id,
                mode=ods.ValueMatrixRequestStruct.ModeEnum.MO_CALCULATED,
                columns=[ods.ValueMatrixRequestStruct.ColumnItem(name="*")],
                attributes=lc_attributes,
                values_start=0,
                values_limit=10,
            )
        )
        assert 1 == len(lc_values_dms.matrices)
        with open(
            os.path.join(tempfile.gettempdir(), f"{test_name}_lc_values.proto.json"), "w", encoding="utf-8"
        ) as json_file:
            json_file.write(MessageToJson(lc_values_dms))


def test_values_valuematrix_read_storage(request: FixtureRequest):
    test_name = request.node.name
    with __create_con_i() as con_i:
        sm_df = con_i.query_data(
            {
                "AoSubmatrix": {"name": "Profile_02"},
                "$attributes": {"id": 1, "number_of_rows": 1},
                "$options": {"$rowlimit": 1},
            }
        )
        assert sm_df.empty is False
        sm_id: int = sm_df.iat[0, 0]
        sm_number_of_rows: int = sm_df.iat[0, 1]
        logging.getLogger().info("Submatrix Id: %s, number_of_rows:%s", sm_id, sm_number_of_rows)
        assert sm_id is not None
        assert sm_number_of_rows is not None

        lc_entity = con_i.mc.entity_by_base_name("AoLocalColumn")
        assert lc_entity is not None

        lc_info_df = con_i.query_data(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "name": 1, "sequence_representation": 1, "independent": 1},
            }
        )
        assert lc_info_df.empty is False

        lc_names = lc_info_df.iloc[:, 1].tolist()
        lc_columns = [ods.ValueMatrixRequestStruct.ColumnItem(name=lc_name) for lc_name in lc_names]

        lc_attributes = [
            con_i.mc.attribute_by_base_name(lc_entity, "name").name,
            con_i.mc.attribute_by_base_name(lc_entity, "values").name,
            con_i.mc.attribute_by_base_name(lc_entity, "generation_parameters").name,
        ]

        lc_info_dms = con_i.data_read_jaquel(
            {
                "AoLocalColumn": {"submatrix": sm_id},
                "$attributes": {"id": 1, "name": 1, "sequence_representation": 1, "independent": 1},
            }
        )
        assert 1 == len(lc_info_dms.matrices)
        with open(
            os.path.join(tempfile.gettempdir(), f"{test_name}_lc_info.proto.json"), "w", encoding="utf-8"
        ) as json_file:
            json_file.write(MessageToJson(lc_info_dms))

        value_matrix_request = ods.ValueMatrixRequestStruct(
            aid=con_i.mc.entity_by_base_name("AoSubmatrix").aid,
            iid=sm_id,
            mode=ods.ValueMatrixRequestStruct.ModeEnum.MO_STORAGE,
            columns=lc_columns,
            attributes=lc_attributes,
            values_start=0,
            values_limit=10,
        )
        value_matrix_request_str = MessageToJson(value_matrix_request, indent=None)
        logging.getLogger().info(value_matrix_request_str)
        lc_values_dms = con_i.valuematrix_read(value_matrix_request)
        assert 1 == len(lc_values_dms.matrices)
        with open(
            os.path.join(tempfile.gettempdir(), f"{test_name}_lc_values.proto.json"), "w", encoding="utf-8"
        ) as json_file:
            json_file.write(MessageToJson(lc_values_dms))


def test_security_property_returns_security_instance(monkeypatch):
    class DummySecurity:
        __slots__ = ("con_i",)

        def __init__(self, con_i):
            object.__setattr__(self, "con_i", con_i)

        def __setattr__(self, key, value):
            raise AttributeError(f"{self.__class__.__name__} is immutable")

    # Patch Security to DummySecurity
    import odsbox.con_i

    monkeypatch.setattr(odsbox.con_i, "Security", DummySecurity)

    with __create_con_i() as con_i:
        # Remove cached security to force property to create it
        con_i.__security = None
        security = con_i.security
        assert isinstance(security, DummySecurity)
        # Should return the same instance if called again
        assert security is con_i.security


def test_security_property_raises_if_no_session():
    with __create_con_i() as con_i:
        # Simulate closed session
        con_i._ConI__session = None  # type: ignore
        try:
            _ = con_i.security
            raise AssertionError("Expected ValueError when session is None")
        except ValueError as e:
            assert "No open session" in str(e)


def test_security_read():
    with __create_con_i() as con_i:
        security_read_request = ods_security.SecurityReadRequest(
            data_object_type=ods_security.DataObjectTypeEnum.DOT_APPLICATION_ELEMENT,
            application_element=ods_security.SecurityReadRequest.DataObjectApplicationElement(
                aid=con_i.mc.entity_by_base_name("AoTest").aid
            ),
        )
        security_info = con_i.security.security_read(security_read_request)
        assert isinstance(security_info, ods_security.SecurityInfo)


def test_security_level():
    with __create_con_i() as con_i:
        entity = con_i.mc.entity_by_base_name("AoTest")
        security_level = Security.Level(entity.security_level)
        assert isinstance(security_level, Security.Level)
        assert Security.Level.ELEMENT not in security_level

    assert 7 == int(Security.Level.ELEMENT | Security.Level.INSTANCE | Security.Level.ATTRIBUTE)


def test_do_not_load_model():
    """Test that ConI can be created without loading the model"""
    with __create_con_i(False) as con_i:
        with pytest.raises(ValueError):
            _ = con_i.mc
        with pytest.raises(ValueError):
            _ = con_i.model()
        # update model
        assert con_i.model_read() is not None
        # Access the cached model
        assert con_i.mc is not None
        con_i.model()



"""helps working with the ASAM ODS model"""

from __future__ import annotations

import logging

import odsbox.proto.ods_pb2 as ods


class ModelCache:
    """
    The ods.Model object returned from ods server needs some utilities to work with it.
    This cache functionality useful for daily work.
    """

    __model: ods.Model
    __log: logging.Logger = logging.getLogger(__name__)

    def __init__(self, model: ods.Model):
        self.__model = model

    def model(self) -> ods.Model:
        """
        Get the attached ASAM ODS model

        :return ods.Model: The used model.
        """
        return self.__model

    def aid(self, entity_or_name: str | ods.Model.Entity) -> int:
        """
        Determine the application element id of an entity by its name.

        :param str entity_or_name: entity object or case sensitive application name to lookup
        :raises ValueError: If the entity does not exist.
        :return int: The ApplicationElementId of the entity.
        """
        return self.__entity(entity_or_name).aid

    def entity(self, entity_name: str) -> ods.Model.Entity:
        """
        Get the entity name.

        :param str entity_name: case insensitive name of an entity.
        :raises ValueError: If the entity does not exist.
        """
        entity = self.__model.entities.get(entity_name)
        if entity is not None:
            return entity
        name_casefold = entity_name.casefold()
        for key, entity in self.__model.entities.items():
            if key.casefold() == name_casefold:
                return entity
        raise ValueError(f"No entity named '{entity_name}' found.")

    def entity_by_base_name(self, entity_base_name: str) -> ods.Model.Entity:
        """
        Get the entity by its base name.

        :param str entity_base_name: case insensitive name of the base model element.
        :raises ValueError: If the entity does not exist.
        """
        name_casefold = entity_base_name.casefold()
        for _, entity in self.__model.entities.items():
            if name_casefold == entity.base_name.casefold():
                return entity
        raise ValueError(f"No entity derived from base type '{entity_base_name}' found.")

    def entity_by_aid(self, aid: int) -> ods.Model.Entity:
        """
        Get the entity by its ApplicationElementId(aid).

        :param int aid: ApplicationElementId of an entity to lookup.
        :raises ValueError: If the entity does not exist.
        :return ods.Model.Entity: Entity corresponding to given aid.
        """
        for _, entity in self.__model.entities.items():
            if aid == entity.aid:
                return entity
        raise ValueError(f"No entity found with aid '{aid}'.")

    def attribute_no_throw(
        self, entity_or_name: str | ods.Model.Entity, application_or_base_name: str
    ) -> ods.Model.Attribute | None:
        """
        This is a convenience method to find an attribute. It will first check for
        an attribute with the given application name and afterwards check for an
        attribute with the given base name.

        :param str | ods.Model.Entity entity_or_name: entity or case insensitive name of an entity
        :param str application_or_base_name: case insensitive name to lookup
        :raises ValueError: If entity does not exist.
        :return ods.Model.Attribute | None: The found attribute or None.
        """
        entity = self.__entity(entity_or_name)
        attribute = entity.attributes.get(application_or_base_name)
        if attribute is not None:
            return attribute
        name_casefold = application_or_base_name.casefold()
        for _, attribute in entity.attributes.items():
            if attribute.name.casefold() == name_casefold or attribute.base_name.casefold() == name_casefold:
                return attribute
        return None

    def attribute(self, entity_or_name: str | ods.Model.Entity, application_or_base_name: str) -> ods.Model.Attribute:
        """
        This is a convenience method to find an attribute. It will first check for
        an attribute with the given application name and afterwards check for an
        attribute with the given base name.

        :param str | ods.Model.Entity entity_or_name: entity or case insensitive name of an entity
        :param str application_or_base_name: case insensitive name to lookup
        :raises ValueError: If attribute does not.
        :return ods.Model.Attribute: The found attribute.
        """
        entity = self.__entity(entity_or_name)
        attribute = self.attribute_no_throw(entity, application_or_base_name)
        if attribute is not None:
            return attribute
        raise ValueError(
            f"'{entity.name}' has no attribute named '{application_or_base_name}' as base or application name."
        )

    def attribute_by_base_name(
        self, entity_or_name: str | ods.Model.Entity, attribute_base_name: str
    ) -> ods.Model.Attribute:
        """
        Get the attribute by base name.

        :param str | ods.Model.Entity entity_or_name: entity object or case sensitive application name to lookup
        :param str attribute_base_name: case insensitive name of the base model element.
        :raises ValueError: If the attribute does not exist.
        :return ods.Model.Attribute: Corresponding attribute.
        """
        entity = self.__entity(entity_or_name)
        attributes = entity.attributes
        for _, attribute in attributes.items():
            if attribute_base_name.casefold() == attribute.base_name.casefold():
                return attribute
        raise ValueError(f"Entity '{entity.name}' does not have attribute derived from '{attribute_base_name}'.")

    def relation_no_throw(
        self, entity_or_name: str | ods.Model.Entity, application_or_base_name: str
    ) -> ods.Model.Relation | None:
        """
        This is a convenience method to find a relation. It will first check for
        a relation with the given application name and afterwards check for a
        relation with the given base name.

        :param str | ods.Model.Entity entity_or_name: entity or case insensitive name of an entity
        :param str application_or_base_name: case insensitive name to lookup
        :raises ValueError: If entity does not exist.
        :return ods.Model.Relation | None: The relation or None if it does not exist.
        """
        entity = self.__entity(entity_or_name)
        relation = entity.relations.get(application_or_base_name)
        if relation is not None:
            return relation
        # retry case insensitive
        name_casefold = application_or_base_name.casefold()
        for _, relation in entity.relations.items():
            if relation.name.casefold() == name_casefold or relation.base_name.casefold() == name_casefold:
                return relation
        return None

    def relation(self, entity_or_name: str | ods.Model.Entity, application_or_base_name: str) -> ods.Model.Relation:
        """
        This is a convenience method to find a relation. It will first check for
        a relation with the given application name and afterwards check for a
        relation with the given base name.

        :param str | ods.Model.Entity entity_or_name: entity or case insensitive name of an entity
        :param str application_or_base_name: case insensitive name to lookup
        :raises ValueError: If relation does not exist.
        :return ods.Model.Relation: The found relation.
        """
        entity = self.__entity(entity_or_name)
        relation = self.relation_no_throw(entity, application_or_base_name)
        if relation is not None:
            return relation
        raise ValueError(
            f"'{entity.name}' has no relation named '{application_or_base_name}' as base or application name."
        )

    def relation_by_base_name(
        self, entity_or_name: str | ods.Model.Entity, relation_base_name: str
    ) -> ods.Model.Relation:
        """
        Get the relation by base name.

        :param str | ods.Model.Entity entity_or_name: entity object or case sensitive application name to lookup
        :param str relation_base_name: case insensitive name of the base model element.
        :raises ValueError: If the relation does not exist.
        :return ods.Model.Relation: Corresponding relation.
        """
        entity = self.__entity(entity_or_name)
        relations = entity.relations
        for _, relation in relations.items():
            if relation_base_name.casefold() == relation.base_name.casefold():
                return relation
        raise ValueError(f"Entity '{entity.name}' does not have relation derived from '{relation_base_name}'.")

    def enumeration(self, enumeration_name: str) -> ods.Model.Enumeration:
        """
        Get enumeration by its name.

        :param str enumeration_name: case insensitive name of the application model enumeration.
        :raises ValueError: If the enumeration does not exist.
        :return ods.Model.Enumeration: Corresponding enumeration.
        """
        enumeration = self.__model.enumerations.get(enumeration_name)
        if enumeration is not None:
            return enumeration
        name_casefold = enumeration_name.casefold()
        for key, enumeration in self.__model.enumerations.items():
            if key.casefold() == name_casefold:
                return enumeration
        raise ValueError(f"Enumeration '{enumeration_name}' does not exist in data model.")

    def enumeration_value_to_key(self, enumeration_or_name: str | ods.Model.Enumeration, lookup_value: int) -> str:
        """
        Convert an enumeration value into its string representation.

        :param str | ods.Model.Enumeration enumeration_or_name: ods enumeration or its case insensitive name
        :param int lookup_value: integer value to check
        :raises ValueError: If the enumeration does not exist or does not contain value.
        :return str: String representation of int value.
        """
        enumeration = self.__enumeration(enumeration_or_name)
        for key, value in enumeration.items.items():
            if value == lookup_value:
                return key
        raise ValueError(f"Enumeration '{enumeration.name}' does not contain the int value '{lookup_value}'.")

    def enumeration_key_to_value(self, enumeration_or_name: str | ods.Model.Enumeration, lookup_key: str) -> int:
        """
        Convert an enumeration integer value into its string representation.

        :param str | ods.Model.Enumeration enumeration_or_name: ods enumeration or its case insensitive name
        :param str lookup_key: case insensitive string key value to check
        :raises ValueError: If the enumeration does not exist or does not contain the key.
        :return str: Int representation of string value.
        """
        enumeration = self.__enumeration(enumeration_or_name)
        if lookup_key in enumeration.items:
            return enumeration.items[lookup_key]
        name_casefold = lookup_key.casefold()
        for key, value in enumeration.items.items():
            if key.casefold() == name_casefold:
                return value
        raise ValueError(f"Enumeration '{enumeration.name}' does not contain the key '{lookup_key}'.")

    def __entity(self, entity_or_name: str | ods.Model.Entity) -> ods.Model.Entity:
        if isinstance(entity_or_name, ods.Model.Entity):
            return entity_or_name
        return self.entity(entity_or_name)

    def __enumeration(self, enumeration_or_name: str | ods.Model.Enumeration) -> ods.Model.Enumeration:
        if isinstance(enumeration_or_name, ods.Model.Enumeration):
            return enumeration_or_name
        return self.enumeration(enumeration_or_name)



"""Integration test for ASAM ODS session"""

import requests
from odsbox.con_i import ConI

import pytest
import os
from unittest import mock
import tempfile
from odsbox.proto.ods_pb2 import FileIdentifier


@pytest.fixture
def con_i():
    return ConI("https://docker.peak-solution.de:10032/api", ("Demo", "mdm"))


def test_file_access_download_success(con_i):
    file_identifier = FileIdentifier(aid=4711, iid=1)
    target_file = "test_download.bin"

    with mock.patch.object(con_i, "file_access", return_value="http://example.com/file"):
        with mock.patch.object(con_i._ConI__session, "get") as mock_get:
            mock_response = mock.Mock()
            mock_response.content = b"file content"
            mock_response.headers = {"Content-Disposition": "attachment; filename=test_download.bin"}
            mock_get.return_value = mock_response

            downloaded_file = con_i.file_access_download(file_identifier, target_file, overwrite_existing=True)

            assert downloaded_file == target_file
            assert os.path.exists(target_file)
            with open(target_file, "rb") as f:
                assert f.read() == b"file content"

            os.remove(target_file)


def test_file_access_download_success_2(con_i):
    file_identifier = FileIdentifier(aid=4711, iid=1)
    target_file = "test_download.bin"

    with mock.patch.object(con_i, "file_access", return_value="http://example.com/file"):
        with mock.patch.object(con_i._ConI__session, "get") as mock_get:
            mock_response = mock.Mock()
            mock_response.content = b"file content"
            mock_response.headers = {"Content-Disposition": 'attachment; filename="test_download.bin"'}
            mock_get.return_value = mock_response

            downloaded_file = con_i.file_access_download(file_identifier, target_file, overwrite_existing=True)

            assert downloaded_file == target_file
            assert os.path.exists(target_file)
            with open(target_file, "rb") as f:
                assert f.read() == b"file content"

            os.remove(target_file)


def test_file_access_download_success_3(con_i):
    with tempfile.TemporaryDirectory() as temp_dir:
        file_identifier = FileIdentifier(aid=4711, iid=1)
        target_file = os.path.join(temp_dir, "test_download.bin")

        with mock.patch.object(con_i, "file_access", return_value="http://example.com/file"):
            with mock.patch.object(con_i._ConI__session, "get") as mock_get:
                mock_response = mock.Mock()
                mock_response.content = b"file content"
                mock_response.headers = {"Content-Disposition": 'attachment; filename="test_download.bin"'}
                mock_get.return_value = mock_response

                downloaded_file = con_i.file_access_download(file_identifier, temp_dir, overwrite_existing=True)

                assert downloaded_file == target_file
                assert os.path.exists(target_file)
                with open(target_file, "rb") as f:
                    assert f.read() == b"file content"


def test_file_access_download_success_4(con_i):
    with tempfile.TemporaryDirectory() as temp_dir:
        file_identifier = FileIdentifier(aid=4711, iid=1)
        target_file = os.path.join(temp_dir, "download.bin")

        with mock.patch.object(con_i, "file_access", return_value="http://example.com/file"):
            with mock.patch.object(con_i._ConI__session, "get") as mock_get:
                mock_response = mock.Mock()
                mock_response.content = b"file content"
                mock_response.headers = {}
                mock_get.return_value = mock_response

                downloaded_file = con_i.file_access_download(file_identifier, temp_dir, overwrite_existing=True)

                assert downloaded_file == target_file
                assert os.path.exists(target_file)
                with open(target_file, "rb") as f:
                    assert f.read() == b"file content"


def test_file_access_download_no_overwrite(con_i):
    file_identifier = FileIdentifier(aid=4711, iid=1)
    target_file = "test_download_no_overwrite.bin"

    with open(target_file, "wb") as f:
        f.write(b"existing content")

    with mock.patch.object(con_i, "file_access", return_value="http://example.com/file"):
        with mock.patch.object(con_i._ConI__session, "get") as mock_get:
            mock_response = mock.Mock()
            mock_response.content = b"new file content"
            mock_response.headers = {"Content-Disposition": "attachment; filename=test_download_no_overwrite.bin"}
            mock_get.return_value = mock_response

            with pytest.raises(FileExistsError):
                con_i.file_access_download(file_identifier, target_file, overwrite_existing=False)

    os.remove(target_file)


def test_file_access_download_no_session(con_i):
    file_identifier = FileIdentifier(aid=4711, iid=1)
    target_file = "test_download_no_session.bin"

    con_i._ConI__session = None

    with pytest.raises(ValueError, match="No open session!"):
        con_i.file_access_download(file_identifier, target_file)


def test_file_access_upload_success(con_i):
    # Create a temporary file to upload
    with tempfile.NamedTemporaryFile(delete=False) as temp_file:
        temp_file.write(b"Test content")
        temp_file_path = temp_file.name

    file_identifier = FileIdentifier(aid=4711, iid=1)

    with mock.patch.object(con_i, "file_access", return_value="http://example.com/upload"):
        with mock.patch.object(con_i._ConI__session, "put") as mock_post:
            mock_response = mock.Mock()
            mock_response.status_code = 200
            mock_post.return_value = mock_response

            con_i.file_access_upload(file_identifier, temp_file_path)

            mock_post.assert_called_once()
            assert mock_post.call_args[0][0] == "http://example.com/upload"

    os.remove(temp_file_path)


def test_file_access_upload_failure(con_i):
    # Create a temporary file to upload
    with tempfile.NamedTemporaryFile() as temp_file:
        temp_file.write(b"Test content")
        temp_file_path = temp_file.name

        file_identifier = FileIdentifier(aid=4711, iid=1)

        with mock.patch.object(con_i, "file_access", return_value="http://example.com/upload"):
            with mock.patch.object(con_i._ConI__session, "put") as mock_post:
                mock_response = mock.Mock()
                mock_response.status_code = 500
                mock_response.headers = {}
                mock_response.raise_for_status.side_effect = requests.HTTPError("Failed to upload file")
                mock_post.return_value = mock_response

                with pytest.raises(requests.HTTPError, match="Failed to upload file"):
                    con_i.file_access_upload(file_identifier, temp_file_path)


def test_file_access_delete_success(con_i):
    file_identifier = FileIdentifier(aid=4711, iid=1)

    with mock.patch.object(con_i, "file_access", return_value="http://example.com/delete"):
        with mock.patch.object(con_i._ConI__session, "delete") as mock_delete:
            mock_response = mock.Mock()
            mock_response.status_code = 200
            mock_delete.return_value = mock_response

            con_i.file_access_delete(file_identifier)

            mock_delete.assert_called_once()
            assert mock_delete.call_args[0][0] == "http://example.com/delete"


def test_file_access_delete_failure(con_i):
    file_identifier = FileIdentifier(aid=4711, iid=1)

    with mock.patch.object(con_i, "file_access", return_value="http://example.com/delete"):
        with mock.patch.object(con_i._ConI__session, "delete") as mock_delete:
            mock_response = mock.Mock()
            mock_response.status_code = 404
            mock_response.headers = {}
            mock_response.raise_for_status.side_effect = requests.HTTPError("File not found")
            mock_delete.return_value = mock_response

            with pytest.raises(requests.HTTPError, match="File not found"):
                con_i.file_access_delete(file_identifier)
